#include <cstddef>
using std::size_t;

#include <cstdlib>
using std::exit;

#include <cassert>

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;

#include <fstream>
using std::ofstream;

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include "mtrand.hpp"
mtrand Rand(1975);

const size_t eighthL = 50;
const size_t quartL = 2*eighthL;
const size_t halfL = 2*quartL;
const size_t L = 2*halfL;
char fluid[L][L];

// storage for coarse-graining
long int vx[halfL][halfL];
long int vy[halfL][halfL];
unsigned long int n[halfL][halfL];

const int delay = 50; // milliseconds

double pos(size_t i) { return 2.0*i/L-1.0; }

void display()
{
	glClear(GL_COLOR_BUFFER_BIT);
	for (size_t i = 0; i < L; ++i)
		for (size_t j = 0; j < L; ++j)
			if (fluid[i][j] != 0)
			{
				if (fluid[i][j] == 1)
					glColor3f(1.0, 1.0, 1.0); 
				else
					glColor3f(1.0, 0.5, 0.25); 
				glBegin(GL_POLYGON);
					glVertex2f(pos(i),pos(j));
					glVertex2f(pos(i+1),pos(j));
					glVertex2f(pos(i+1),pos(j+1));
					glVertex2f(pos(i),pos(j+1));
				glEnd();
			}
	glutSwapBuffers();
}

void keyboard(unsigned char key, int x, int y)
{
	if (key == 'q')
		exit(0);
}

int setup;
double concentration;
bool check_conservation_laws = true;
void initialize_grid(void)
{
	switch(setup)
	{
	case 0: // fluid 
		for (size_t i = 0; i < L; ++i)
			for (size_t j = 0; j < L; ++j)
				fluid[i][j] = (Rand() < concentration ? 1 : 0);
		break;

	case 1: // fluid with a low density hole
		check_conservation_laws=false;
		for (size_t i = 0; i < L; ++i)
			for (size_t j = 0; j < L; ++j)
			{
				const double x = double(i) - double(quartL);
				const double y = double(j) - double(quartL);
				const double c = (x*x + y*y < 0.5*quartL*quartL ? 0.1*concentration : concentration);				
				fluid[i][j] = (Rand() < c ? 1 : 0);
			}
		break;
		
	case 2: // fluid in a container
		check_conservation_laws=false;
		for (size_t i = 3; i < halfL; ++i)
			for (size_t j = 3; j < L-3; ++j)
				fluid[i][j] = (Rand() < concentration ? 1 : 0);
		
			for (size_t i = 0; i < L; ++i)
		{
			fluid[i][0] = 0;
			fluid[i][L-1] = 0;
			fluid[0][i] = 0;
			fluid[L-1][i] = 0;
		}

		for (size_t i = 1; i < L-1; ++i)
		{
			fluid[i][1] = fluid[i][2] = 2;
			fluid[i][L-2] = fluid[i][L-3] = 2;
			fluid[1][i] = fluid[2][i] = 2;
			fluid[L-2][i] = fluid[L-3][i] = 2;
			if (i < L/3-2 or i > L/3+2)
				fluid[halfL][i] = fluid[halfL+1][i] = 2;
		}

		break;

	default:
		assert(false);
		break;
	}
		
	for (size_t i = 0; i < halfL; ++i)
		for (size_t j = 0; j < halfL; ++j)
			vx[i][j] = vy[i][j] = 0;
}

void update(char &TL, char &TR, char &BL, char &BR)
{
	char occupancy = (TL == 1) + (TR == 1) + (BL == 1) + (BR == 1);
    if (occupancy == 1)
	{
		if (TL == 1 and BR != 2) { TL = 0; BR = 1; }
		else if (TL == 1 and BL == 2 and TR != 2) { TL = 0; TR = 1; }
		else if (TL == 1 and TR == 2 and BL != 2) { TL = 0; BL = 1; }
		else if (TR == 1 and BL != 2) { TR = 0; BL = 1; }
		else if (TR == 1 and TL == 2 and BR != 2) { TR = 0; BR = 1; }
		else if (TR == 1 and BR == 2 and TL != 2) { TR = 0; TL = 1; }
		else if (BR == 1 and TL != 2) { BR = 0; TL = 1; }
		else if (BR == 1 and BL == 2 and TR != 2) { BR = 0; TR = 1; }
		else if (BR == 1 and TR == 2 and BL != 2) { BR = 0; BL = 1; }
		else if (BL == 1 and TR != 2) { BL = 0; TR = 1; }
		else if (BL == 1 and TL == 2 and BR != 2) { BL = 0; BR = 1; }
		else if (BL == 1 and BR == 2 and TL != 2) { BL = 0; TL = 1; }
	}
	else if (occupancy == 3)
	{
		if (TL == 0 and TL != 2) { TL = 1; BR = 0; }
		else if (TR == 0 and TR != 2) { TR = 1; BL = 0; }
		else if (BR == 0 and BR != 2) { BR = 1; TL = 0; }
		else if (BL == 0 and BL != 2) { BL = 1; TR = 0; }			
	}
	else if (occupancy == 2)
	{
		if (TL == 1 and BR == 1 and BL != 2 and TR != 2) { TL = BR = 0; BL = TR = 1; }
		else if (TR == 1 and BL == 1 and BR != 2 and TL != 2) { TR = BL = 0; BR = TL = 1; }
		else if (TL == 1 and BL == 1 and TR != 2 and BR != 2) { TL = BL = 0; TR = BR = 1; }
		else if (TL == 1 and TR == 1 and BL != 2 and BR != 2) { TL = TR = 0; BL = BR = 1; }
		else if (TR == 1 and BR == 1 and TL != 2 and BL != 2) { TR = BR = 0; TL = BL = 1; }
		else if (BL == 1 and BR == 1 and TL != 2 and TR != 2) { BL = BR = 0; TL = TR = 1; }
	}
}	

void verify(void) // conservation of mass/energy
{
	static bool first_time = true;
	static int last_count = 0;

	int count = 0;
	for (size_t i = 0; i < L; ++i)
		for (size_t j = 0; j < L; ++j)
			count += fluid[i][j];
	
	if (!first_time and last_count != count) cerr << "WARNING: particle number not conserved!" << endl;
	
	last_count = count;
	
	static int last_momentum_x = 0;
	static int last_momentum_y = 0;

	int momentum_x = 0;
	int momentum_y = 0;
	for (size_t i = 0; i < L; i+=2)
		for (size_t j = 0; j < L; j+=2)
		{
			if (fluid[i][j] == 1) { --momentum_x; --momentum_y; }
			if (fluid[i+1][j] == 1) { ++momentum_x; --momentum_y; }
			if (fluid[i+1][j+1] == 1) { ++momentum_x; ++momentum_y; }
			if (fluid[i][j+1] == 1) { --momentum_x; ++momentum_y; }
		}

	//cout << momentum_x << "\t" << momentum_y << endl;
	if (!first_time and (last_momentum_x != momentum_x or last_momentum_y != momentum_y))
		cerr << "WARNING: momentum not conserved!" << endl;

	last_momentum_x = momentum_x; 
	last_momentum_y = momentum_y;
	first_time = false;
}

void measure(void)
{
	for (size_t i = 0; i < L; i += 2)
		for (size_t j = 0; j < L; j += 2)
		{
			if (fluid[i][j] == 1) { ++vx[i/2][j/2]; ++vy[i/2][j/2]; ++n[i/2][j/2]; }
			if (fluid[i+1][j] == 1) { --vx[i/2][j/2]; ++vy[i/2][j/2]; ++n[i/2][j/2]; }
			if (fluid[i+1][j+1] == 1) { --vx[i/2][j/2]; --vy[i/2][j/2]; ++n[i/2][j/2]; }
			if (fluid[i][j+1] == 1) { ++vx[i/2][j/2]; --vy[i/2][j/2]; ++n[i/2][j/2]; }			
		}
}

size_t num_measurements = 0;
void write_data(void)
{
	cout << "Writing coarse-grained data" << endl;
	
	for (size_t i = 0; i < quartL; ++i)
		for (size_t j = 0; j < quartL; ++j)
		{
			vx[i][j] = vx[2*i][2*j] + vx[2*i+1][2*j] + vx[2*i][2*j+1] + vx[2*i+1][2*j+1];
			vy[i][j] = vy[2*i][2*j] + vy[2*i+1][2*j] + vy[2*i][2*j+1] + vy[2*i+1][2*j+1];
			n[i][j] = n[2*i][2*j] + n[2*i+1][2*j] + n[2*i][2*j+1] + n[2*i+1][2*j+1];
		}
	
	for (size_t i = 0; i < eighthL; ++i)
		for (size_t j = 0; j < eighthL; ++j)
		{
			vx[i][j] = vx[2*i][2*j] + vx[2*i+1][2*j] + vx[2*i][2*j+1] + vx[2*i+1][2*j+1];
			vy[i][j] = vy[2*i][2*j] + vy[2*i+1][2*j] + vy[2*i][2*j+1] + vy[2*i+1][2*j+1];
			n[i][j] = n[2*i][2*j] + n[2*i+1][2*j] + n[2*i][2*j+1] + n[2*i+1][2*j+1];
		}
	
	ofstream fout("vfield.dat");
	for (size_t i = 0; i < eighthL; ++i)
		for (size_t j = 0; j < eighthL; ++j)
			fout << i << "\t" << j << "\t" << double(vx[i][j])/num_measurements << "\t" << double(vy[i][j])/num_measurements << endl;
	fout.close();
	
	fout.open("density.dat");
	for (size_t j = 0; j < eighthL; ++j)
	{
		for (size_t i = 0; i < eighthL; ++i)
			fout << double(n[i][j])/num_measurements << "\t";
		fout << endl;
	}
	fout.close();
}

size_t simulation_clock = 0;
bool start_measurement = false;
void sweep_grid(void)
{
	const size_t offset = simulation_clock%2;
	for (size_t i = offset; i < L; i += 2)
		for (size_t j = offset; j < L; j += 2)
			update(fluid[i][j],fluid[(i+1)%L][j],fluid[i][(j+1)%L],fluid[(i+1)%L][(j+1)%L]);
	
	++simulation_clock;	
	if (simulation_clock%2==0)
	{
		if (check_conservation_laws) verify();
		
		if (!start_measurement and setup==2 and simulation_clock > 50*L) 
		{
			cout << "Begin measurement taking" << endl;
			start_measurement = true;
		}
		
		if (start_measurement) 
		{
			++num_measurements;
			measure();
			
			if (num_measurements > 75*L) 
			{
				write_data();
				exit(0);
			}
		}		
	}
}

void animate(int)
{
	sweep_grid();
	glutPostRedisplay();
	glutTimerFunc(delay,animate,0);
}

int main(int argc, char **argv)
{
	if (argc != 3)
	{
		cerr << "Usage: hpp (setup=0,1,2) (0 < concentration < 1)" << endl;
		exit(1);
	}
	setup = atoi(argv[1]);
	concentration = atof(argv[2]);
	if (setup < 0 or setup > 2) 
	{
		cerr << "Error: `setup` parameter must be 0, 1, or 2" << endl;
		exit(0);
	}
	if (concentration < 0.0 or concentration > 1.0) 
	{
		cerr << "Error: `concentration` parameter must be in the interval [0,1]" << endl;
		exit(0);
	}
	initialize_grid();	

	glutInit(&argc,argv);
	glutInitDisplayMode (GLUT_RGB | GLUT_DOUBLE);
	glutInitWindowSize (800, 800);
	glutInitWindowPosition (100, 100);
	glutCreateWindow("The HPP model lattice gas");
	glutDisplayFunc(display);
	glutKeyboardFunc(keyboard);
	glutTimerFunc(delay,animate,0);
	glutMainLoop();
}

